#!/usr/bin/env python
# Copyright (C) 2009 Stian Selnes <stian.selnes@gmail.com>
# Published under the GNU GPL v3 or later
# Initially based on the bzr-text-checker plugin by Marius Kruger
# See README for instructions on how to use the plugin.

version_info = (0, 1, 1)
plugin_name = 'style-checker'

import style_checker
from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
from bzrlib import builtins
from bzrlib import branch, commands,option
from bzrlib import trace
""")


def pre_commit_hook(local, master, old_revno, old_revid, future_revno,
        future_revid, tree_delta, future_tree):

    if style_checker._skip_style_check:
       return

    new_tree = future_tree
    if local is not None:
        local_branch = local
    else:
        local_branch = master

    # Find root path of working tree.
    # This is a hack, there must be a better way
    import os, sys
    base = local_branch.base.split("://", 1)
    if base[0] == "file":
        root_dir = base[1]
    elif len(base) == 1:
        root_dir = base[0]
    # Remove first '/' on Windows
    if sys.platform == 'win32':
        if root_dir[0] == '/':
            root_dir = root_dir[1:]

    root_dir = os.path.normpath (root_dir)

    config = style_checker.StyleCheckerConfig (local_branch)
    if not config.found:
        return
    checker = style_checker.StyleChecker (root_dir, config)
    lister = style_checker.FileLister (root_dir, config.file_filters)
    checker.check (lister.list_changed_files (tree_delta))

    trace.note ("Checked code style of %d files, %d violations."
                % (checker.num_checked, checker.num_violations))



old_commit = commands.get_cmd_object("commit",True).__class__
class cmd_commit(old_commit):
    takes_options = old_commit.takes_options + \
        [option.Option ("skip-style-check",
                        help="Don't perform checks on code style.")]

    __doc__ = old_commit.__doc__

    def run(self, skip_style_check=False, **kwargs):
        old_value = style_checker._skip_style_check
        try:
            style_checker._skip_style_check = skip_style_check
            old_commit.run(self, **kwargs)
        finally:
            style_checker._skip_style_check = old_value



class cmd_style_check (commands.Command):
    """Modify files to conform to a given code style.

    An external program will be used to check the code style for the files.
    Only the features of that program limits what you can fix by running
    this command. See README file for the plugin for help on how to configure
    the style check."""

    takes_args = ['file*']
    takes_options = [option.Option ("all", help="Do style checking on all files "
                                    "matching configured pattern, not only "
                                    "changed files."),
                     option.Option ("no-backup", help="Don't save a copy of the "
                                    "files before correcting them.")]


    def run (self, all=False, file_list=None, no_backup=False):

        from bzrlib.workingtree import WorkingTree
        wt =  WorkingTree.open_containing()[0]
        root_dir = wt.abspath("")
        config = style_checker.StyleCheckerConfig (wt.branch)
        if not config.found:
            trace.warning ("Could not find configuration for style checking this location.\n"
                           "See \"bzr help style-check\" for help")
            return
        lister = style_checker.FileLister (root_dir, config.file_filters)

        if all:
            file_list = lister.list_all_files (wt)
        elif file_list is not None:
            file_list = lister.list_files (file_list, wt)
        else:
            delta = wt.changes_from (wt.basis_tree ())
            file_list = lister.list_changed_files (delta)

        checker = style_checker.StyleChecker (root_dir, config)
        checker.fix (file_list, no_backup)

        trace.note ("Checked code style of %d files, %d violations, %d fixed."
                    % (checker.num_checked, checker.num_violations, checker.num_fixed))


branch.Branch.hooks.install_named_hook('pre_commit',
    pre_commit_hook, 'style-checker: Check code style of files')
commands.register_command(cmd_commit, decorate=True)
commands.register_command(cmd_style_check, decorate=True)
