#!/usr/bin/env python
# Copyright (C) 2008 Stian Selnes <stianse@gmail.com>
# Published under the GNU GPL v3 or later
# Initially based on the bzr-text-checker plugin by Marius Kruger


from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os, subprocess, shutil
import tempfile, string
from bzrlib import (
    errors,
    globbing,
    trace,
    )
""")
import _lib.filecmp


_skip_style_check = False


class StyleCheckFailed (errors.BzrError):

    _fmt = ("%(msg)s")

    def __init__(self, message):
        errors.BzrError.__init__(self)
        self.msg = "StyleCheckFailed: " + message

class StyleCheckConfigError (errors.BzrError):

    _fmt = ("%(msg)s")

    def __init__(self, message):
        errors.BzrError.__init__(self)
        self.msg = "StyleCheckConfigError: " + message


class StyleCheckCommandFailed(errors.BzrError):

    _fmt = ("%(msg)s")

    def __init__(self, message):
        errors.BzrError.__init__(self)
        self.msg = "StyleCheckCommandFailed: " + message


class StyleCheckerConfig:
    """Class to fetch and store configuration for the plugin. """

    command_template = None
    command_template_string = None
    file_filters = None
    action = None
    found = True

    def __init__ (self, branch):
        """ """
        config = branch.get_config()

        self.command_template_string = config.get_user_option(
            'style_checker_command')
        if self.command_template_string == None:
            self.found = False
            return

        self.command_template = string.Template (self.command_template_string)

        filters = config.get_user_option('style_checker_file_filters')
        if isinstance(filters, unicode) or isinstance(filters, str):
            # if it doesn't have a comma somewhere, get_value returns a
            # unicode value, but we want a list
            filters = [filters,]
        if filters is None:
            filters = []
        self.file_filters = [f for f in filters if len(f) > 0]

        action = config.get_user_option('style_checker_action')
        if action is not None:
            self.action = action
        else:
            self.action = "fail"

class StyleChecker:
    """Class who's task is to check for correct code style in files
    modified/added in a working tree."""

    _config = None
    _root_dir = "."
    num_checked = 0
    num_violations = 0
    num_fixed = 0


    def __init__ (self, root_dir, config):
        """ """
        self._config = config
        self._root_dir = os.path.normpath (root_dir)


    def _backup (self, files):
        """Backup given files. For now, we just copy the files to a directory
        under to root of the working tree. """

        backup_dir = ".style-checker-backup"

        for src in files:
            dst = os.path.join (backup_dir, src)
            dstdir = os.path.dirname (dst)
            if not os.path.exists (dstdir):
                os.makedirs (dstdir)
            shutil.copyfile (src, dst)


    def _check_file (self, path, fix=False):
        """ Checks code style of given file. Returns TRUE if style is OK. If
        FIX is True the file PATH will be overwritten with a fixed version of
        that file."""

        # Write the output of the external code style program to a temporary
        # file and check if there are differences

        tmp_fd, tmp_path = tempfile.mkstemp()
        try:
            cmd = self._config.command_template.substitute (
                ROOT=self._root_dir, IN=path, OUT=tmp_path)
        except (KeyError, ValueError), e:
            msg = "Failed parsing template, check your keywords: `%s'" % \
                self._config.command_template_string
            raise StyleCheckConfigError (msg)

        try:
            retcode = subprocess.call (split_command (cmd))
            if retcode != 0:
                self.num_violations = self.num_violations + 1
                raise StyleCheckCommandFailed (cmd)
        except (OSError, ValueError):
            print "Unexpected error while running command `%s'" % cmd
            raise


        style_ok = _lib.filecmp.cmp (path, tmp_path, shallow=False)
        self.num_checked = self.num_checked + 1
        if not style_ok:
            self.num_violations = self.num_violations + 1

        if not style_ok and fix:
            # Overwrite the original file with the fixed tmp file. Could have
            # just moved the file, but seems like the tmp file can't be deleted
            # at this point on Windows. It's still accessed and we get an
            # annoying error message.
            shutil.copyfile (tmp_path, path)
            self.num_fixed = self.num_fixed + 1
        try: os.remove (tmp_path)
        except OSError: pass

        return style_ok


    def check (self, files):
        """Checks the code style of given files. Warns or raises and
        exception (depending on configuration) on code style violation."""

        all_ok = True
        orig_dir = os.getcwd ()
        os.chdir (self._root_dir)

        for file_path in files:
            try:
                style_ok = self._check_file (file_path)
            except StyleCheckCommandFailed, e:
                trace.warning (e.msg)
                style_ok = False
            except StyleCheckConfigError, e:
                trace.warning (e.msg)
                style_ok = False
                all_ok = False
                break

            if not style_ok:
                trace.warning ("Code style violation in %s" % file_path)
                all_ok = False

        if not all_ok:
            msg = "Some files do not conform to code style"
            if self._config.action == 'warn':
                trace.warning (msg)
            else:
                os.chdir (orig_dir)
                raise StyleCheckFailed (msg)

        os.chdir (orig_dir)


    def fix (self, files, no_backup):
        """Fixes the code style of given files."""

        orig_dir = os.getcwd ()
        os.chdir (self._root_dir)

        if not no_backup:
            self._backup (files)

        for file_path in files:
            try:
                was_ok = self._check_file (file_path, fix=True)
                if not was_ok:
                    trace.note ("Fixed %s" % file_path)
            except StyleCheckCommandFailed, e:
                trace.warning (e.msg)
            except StyleCheckConfigError, e:
                trace.warning (e.msg)
                break

        os.chdir (orig_dir)


class FileLister:
    """Class to create a list of files that matches given criterias """

    _globster = None
    _root_dir = "."

    def __init__ (self, root_dir, file_filters):
        """ """
        self._root_dir = os.path.normpath (root_dir)
        self._globster = globbing.Globster(file_filters)


    def _get_modified (self, delta):
        """ """
        return [path
                for (path, id, kind, text_changed, meta_modified)
                in delta.modified
                if text_changed and kind == "file"]


    def _get_renamed (self, delta, only_modified=False):
        """ """
        return [path
                for (oldpath, path, id, kind, text_changed, meta_modified)
                in delta.renamed
                if ((not only_modified) or (only_modified and text_changed))
                and kind == "file"]


    def _get_added (self, delta):
        """ """
        return [path
                for (path, id, kind)
                in delta.added
                if kind == "file"]


    def list_changed_files (self, delta):
        """List files in TreeDelta that mathces the filter and are changed.
        This includes added and moved files."""

        files = \
            self._get_modified (delta) + \
            self._get_renamed (delta, only_modified=True) + \
            self._get_added (delta)

        files = filter (lambda f: self._globster.match (f), files)
        files = map (lambda f: os.path.normcase (f), files)

        return files


    def list_all_files (self, working_tree):
        """List all files in working tree that matches the filter. """

        working_tree.lock_read ()

        files = working_tree.list_files ()
        files = map (lambda tuple: os.path.normcase (tuple[0]), files)
        files = filter (lambda f: self._globster.match (f), files)

        working_tree.unlock ()

        return files


    def list_files (self, files, working_tree):
        """List given files in a format that conforms to the other methods
        of this class."""
        return map (lambda f: working_tree.relpath (f), files)


def split_command (cmd):
    """Splits a command into an array. The command is split on whitespace
    except when the whitespace is between two " or ' characters."""
    ret = []
    markers = ["'", '"']
    marker = None

    for curr in cmd.split():
        if marker:
            segment += " " + curr
            if curr.endswith (marker):
                ret.append (segment[:-1])
                marker = None
        else:
            if curr[0] in markers:
                marker = curr[0]
                if curr.endswith (marker):
                    ret.append (curr[1:-1])
                    marker = None
                else:
                    segment = curr[1:]
            else:
                ret.append (curr)

    if marker:
        # A marker was not closed. Should not happen, but in case it does we
        # just pretend it as it wasn't there.
        for curr2 in curr.split():
            ret.append (curr2)

    return ret
